/*******************************************************************************
 * $Id: scoreAlignment.cpp,v 1.14 2004/02/28 00:01:31 darling Exp $
 * This file is copyright 2002-2004 Aaron Darling.  All rights reserved.
 * Please see the file called COPYING for licensing, copying, and modification
 * rights.  Redistribution of this file, in whole or in part is prohibited
 * without express permission.
 ******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "libMems/MemSubsets.h"
#include "libMems/MatchList.h"
#include <string>
#include <fstream>
#include <vector>
#include <algorithm>
#include <iostream>
#include "libMems/IntervalList.h"
#include "libGenome/gnFilter.h"
using namespace std;

// basic data structures

/** store a pair of aligned positions and the characters */
typedef struct aligned_coords_s {
	int64 pos1;
	int64 pos2;
	char char1;
	char char2;
} aligned_coords_t;


class AlignedCoordSeqIComparator {
public:
	boolean operator()(const aligned_coords_t& a, const aligned_coords_t& b) const{
		if( abs(a.pos1) == abs(b.pos1) )
			return abs(a.pos2) < abs(b.pos2);
		return abs(a.pos1) < abs(b.pos1);
	}
};


void constructCoordList( uint seqI, uint seqJ, IntervalList& iv_list, vector< aligned_coords_t >& coord_list, vector< gnSequence* >& seq_table ){

	//
	// pre-allocate the vector
	//
	gnSeqI ij_vec_size = 0;
	for( int ivI = 0; ivI < iv_list.size(); ivI++ ){
		GappedAlignment* aln;
		aln = dynamic_cast< GappedAlignment* >( iv_list[ ivI ].matches[ 0 ] );
		if( aln == NULL ){
			throw "Error:  expecting interval to contain a single GappedAlignment";
		}
		ij_vec_size += aln->Length();
	}
	coord_list = vector< aligned_coords_t >( ij_vec_size );

	//
	// fill in the vector with all aligned pairs
	//
	gnSeqI vecI = 0;	// current place in vector
	for( int ivI = 0; ivI < iv_list.size(); ivI++ ){
		GappedAlignment* aln;
		aln = dynamic_cast< GappedAlignment* >( iv_list[ ivI ].matches[ 0 ] );
		if( aln == NULL ){
			throw "Error:  expecting interval to contain a single GappedAlignment";
		}
		int64 pos1 = aln->Start( seqI );
		int64 pos2 = aln->Start( seqJ );
		// if rev. comp then we're starting at the other (right) side
		if( pos1 < 0 )
			pos1 -= aln->Length( seqI ) - 1;
		if( pos2 < 0 )
			pos2 -= aln->Length( seqJ ) - 1;

		for( gnSeqI colI = 0; colI < aln->Length(); colI++ ){
			aligned_coords_t act;
			act.char1 = aln->align_matrix[ seqI ][ colI ];
			act.char2 = aln->align_matrix[ seqJ ][ colI ];
			act.pos1 = act.char1 == '-' ? 0 : pos1;
			act.pos2 = act.char2 == '-' ? 0 : pos2;
			coord_list[ vecI++ ] = act;
			if( act.char1 != '-' )
				pos1++;
			if( act.char2 != '-' )
				pos2++;
		}
	}

	//
	// sort the vector on aligned position
	//
	AlignedCoordSeqIComparator acsc;
	sort( coord_list.begin(), coord_list.end(), acsc );
}


const gnFilter* comp_filter = gnFilter::DNAComplementFilter();

void sanityCheck( string& aln_name, uint seqI, uint seqJ, aligned_coords_t& act, vector< string >& evolved_seqs ){
	if( act.pos1 != 0 ){
		char seqI_char = evolved_seqs[ seqI ][ abs( act.pos1 ) - 1 ];
		if( act.pos1 < 0 )
			seqI_char = comp_filter->Filter( seqI_char );
		if( act.char1 != seqI_char ){
			cerr << "Error: " << aln_name << " character " << act.char1 << " instead of " << seqI_char << " at position " << act.pos1 << endl;
		}
	}

	if( act.pos2 != 0 ){
		char seqJ_char = evolved_seqs[ seqJ ][ abs( act.pos2 ) - 1 ];
		if( act.pos2 < 0 )
			seqJ_char = comp_filter->Filter( seqJ_char );

		if( act.char2 != seqJ_char ){
			cerr << "Error: " << aln_name << " character " << act.char2 << " instead of " << seqJ_char << " at position " << act.pos2 << endl;
		}
	}
}

void compareAlignments( IntervalList& correct, IntervalList& calculated, vector< string >& evolved_seqs, vector< gnSequence* >& seq_table ){
	
	string cor_name = "correct";
	string calc_name = "calculated";
	gnSeqI sp_truepos = 0;
	gnSeqI sp_falsepos = 0;
	gnSeqI sp_trueneg = 0;
	gnSeqI sp_falseneg = 0;
	gnSeqI sp_extraneg = 0;

	// calculate total possible score
	gnSeqI sp_possible = 0;

	uint seqI = 0;
	uint seqJ = 0;

	// now for every entry in the correct alignment, look for a corresponding
	// entry in the calculated alignment.  Do sanity checking while we're at it
	for( seqI = 0; seqI < seq_table.size(); seqI++ ){
		for( seqJ = seqI+1; seqJ < seq_table.size(); seqJ++ ){
			//
			// the correct coordinate matrix is expected to have one entry for
			// each character in seqI, plus an arbitrary number of "zero" entries
			// for other sequences (seqJ > seqI) that couldn't align to anything
			// in seqI
			//
			vector< aligned_coords_t > cor;
			constructCoordList( seqI, seqJ, correct, cor, seq_table );

			//
			// the calculated alignment is expected to account for every residue
			// at least once, either aligning it to another residue or a gap
			//
			vector< aligned_coords_t > calc;
			constructCoordList( seqI, seqJ, calculated, calc, seq_table );

			gnSeqI corI = 0;
			gnSeqI calcI = 0;

			// skip any gaps aligned to gaps
			while( corI < cor.size() && cor[ corI ].pos1 == 0 && cor[ corI ].pos2 == 0 )
				corI++;
			while( calcI < calc.size() && calc[ calcI ].pos1 == 0 && calc[ calcI ].pos2 == 0 )
				calcI++;

			//
			// get thru the unaligned region of seqI (where pos1 is zero)
			//
			while( corI < cor.size() && calcI < calc.size() && 
				cor[ corI ].pos1 == 0 && calc[ calcI ].pos1 == 0 )
			{
				// sanity check the "correct" aligned characters...
				sanityCheck( cor_name, seqI, seqJ, cor[ corI ], evolved_seqs );

				while( calcI < calc.size() && calc[ calcI ].pos1 == 0 &&
					abs(calc[ calcI ].pos2) < abs(cor[ corI ].pos2) )
				{
					// sanity check the aligned characters...
					sanityCheck( calc_name, seqI, seqJ, calc[ calcI ], evolved_seqs );
					// these are all false positives, catch them later
					calcI++;
				}

				uint match_count = 0;
				while( calcI < calc.size() && calc[ calcI ].pos1 == 0 &&
					abs(calc[ calcI ].pos2) == abs(cor[ corI ].pos2) ){
					// sanity check the aligned characters...
					sanityCheck( calc_name, seqI, seqJ, calc[ calcI ], evolved_seqs );
					match_count++;
					calcI++;
				}
				sp_trueneg += match_count > 0 ? 1 : 0;
				sp_extraneg += match_count > 0 ? match_count - 1 : 0;
				// catch false positives later
				corI++;
			}

			// scan thru any remaining false negatives
			// that were in the calculated alignment
			while( calcI < calc.size() && calc[ calcI ].pos1 == 0 ){
				// sanity check the aligned characters...
				sanityCheck( calc_name, seqI, seqJ, calc[ calcI ], evolved_seqs );
				calcI++;
				sp_falseneg++;
			}
			while( corI < cor.size() && cor[ corI ].pos1 == 0 ){
				sanityCheck( cor_name, seqI, seqJ, cor[ corI ], evolved_seqs );
				// this is a false positive...
				corI++;
			}
			
			//
			// now look at the rest of seqI
			//
			for( ; corI < cor.size(); corI++ ){
				// sanity check the "correct" aligned characters
				sanityCheck( cor_name, seqI, seqJ, cor[ corI ], evolved_seqs );

				// if this is an aligned pair it gets counted towards
				// the total possible points in the sum of pairs score
				if( cor[ corI ].pos2 != 0 )
					sp_possible++;

				// make sure calc is up to where cor is...
				while( calcI < calc.size() && (abs(calc[ calcI ].pos1) < abs(cor[ corI ].pos1)) ){
					// sanity check the aligned characters
					sanityCheck( calc_name, seqI, seqJ, calc[ calcI ], evolved_seqs );

					// cor skipped more than one char and is missing something...
					cerr << "correct alignment is missing characters\n";
					calcI++;
				}

				// make sure cor is up to where calc is...
				if( calcI == calc.size() || (abs(calc[ calcI ].pos1) > abs(cor[ corI ].pos1)) ){
					// calc is missing something!
					// assume it's a gap
					cerr << "Warning, calculated alignment missing prediction for seqI " << seqI << " position " << abs(cor[ corI ].pos1) << ".  Assuming gap prediction\n";
					if( cor[ corI ].pos2 == 0 )
						sp_trueneg++;
					else
						sp_falseneg++;
					continue;
				}

				// if all is well, we should have found the corresponding entries in
				// cor and calc
				int cur_count = 0;
				while( calcI < calc.size() && (abs(calc[ calcI ].pos1) == abs(cor[ corI ].pos1)) ){
					// sanity check the aligned characters...
					sanityCheck( calc_name, seqI, seqJ, calc[ calcI ], evolved_seqs );
					
					// check the strands that were aligned
					bool cor_parity = ((cor[ corI ].pos1 > 0 && cor[ corI ].pos2 > 0) ||
										(cor[ corI ].pos1 < 0 && cor[ corI ].pos2 < 0)) ?
										true : false;

					bool calc_parity = ((calc[ calcI ].pos1 > 0 && calc[ calcI ].pos2 > 0) ||
										(calc[ calcI ].pos1 < 0 && calc[ calcI ].pos2 < 0)) ?
										true : false;

					// if they match then increment the sp_score
					if( abs(calc[ calcI ].pos2) == abs(cor[ corI ].pos2) ){
						if( calc[ calcI ].pos2 == 0 )
							sp_trueneg++;	// correct alignment to a gap
						else if( cor_parity == calc_parity )
							sp_truepos++;	// correct positions on correct strand!!
						else
							sp_falsepos++;	// correct positions, wrong strand
					}else{
						if( calc[ calcI ].pos2 == 0 )
							sp_falseneg++;	// incorrect alignment to a gap
						else
							sp_falsepos++;	// incorrectly aligned

					}
					calcI++;
				}
			}
		}
	}
	
	
	double sp_accuracy = (double)sp_truepos / (double)sp_possible;
	cout << "Sum of pairs accuracy: " << sp_accuracy << endl;
	cout << "Sum of pairs error rate: " << (double)sp_falsepos / (double)sp_truepos << endl;
	cout << "The error rate gives the number of incorrect orthology predictions per correct prediction." << endl;
	cout << "trueneg: " << sp_trueneg << "  falseneg: " << sp_falseneg << "  extraneg: " << sp_extraneg << endl;
}

/**
 * program to score alignments
 * reads in a "correct" alignment and a calculated alignment
 * scores the calculated alignment based on the correct one
 */
int main( int argc, char* argv[] ){
	
	if( argc < 3 ){
		cout << "scoreAlignment <correct alignment> <calculated alignment> [evolved sequence file]\n";
		return -1;
	}
	
	boolean debug_mismatches = false;	/**< turns on code to debug mismatches in evolved and aligned base pairs */
	string correct_fname = argv[ 1 ];
	string calculated_fname = argv[ 2 ];
	string evolved_fname;
	if( argc > 3 ){
		debug_mismatches = true;
		evolved_fname = argv[ 3 ];
	}

	ifstream correct_in;
	correct_in.open( correct_fname.c_str() );
	if( !correct_in.is_open() ){
		cerr << "Error opening " << correct_fname << endl;
		return -1;
	}
	ifstream calculated_in;
	calculated_in.open( calculated_fname.c_str() );
	if( !calculated_in.is_open() ){
		cerr << "Error opening " << calculated_fname << endl;
		return -1;
	}
try{
	IntervalList correct_ivs;
	IntervalList calculated_ivs;
	correct_ivs.ReadStandardAlignment( correct_in );
	correct_in.close();
	calculated_ivs.ReadStandardAlignment( calculated_in );
	calculated_in.close();
	gnSequence empty_seq;
	vector< gnSequence* > seq_table( correct_ivs[0].SeqCount(), &empty_seq );
	uint seq_count = seq_table.size();
	const gnFilter* comp_filter = gnFilter::DNAComplementFilter();
	
	gnSequence evolved_gnseqs;
	vector< string > evolved_seqs( seq_count );
	if( debug_mismatches ){
		evolved_gnseqs.LoadSource( evolved_fname );
		for( uint i = 0; i < seq_count; i++ ){
			evolved_seqs[ i ] = evolved_gnseqs.contig( i ).ToString();
		}
	}
	
	compareAlignments( correct_ivs, calculated_ivs, evolved_seqs, seq_table );

	
}catch( gnException& gne ){
	cerr << gne << endl;
}catch( exception& e ){
	cerr << e.what() << endl;
}catch( char const* c ){
	cerr << c << endl;
}

}


